/*
	Image Capture Daemon - a D-BUS service for interfacing to SANE
	Copyright (C) 2007  Donald Straney <burntfuse@gmail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <sane/sane.h>
#include <dbus/dbus.h>

#include "imcapd.h"
#include "button-notify.h"

//the max number of buttons that can be handled
#define MAX_BUTTONS 20

//used to represent one device and its buttons (chained together in linked list)
struct device_button_st;

struct device_button_st {
	SANE_Handle handle;		//the device's SANE handle
	int buttonlist[MAX_BUTTONS];	//a list of the device's button option numbers -
										// first non-button is 0
	struct device_button_st *next;	//pointer to the next device
	struct device_button_st *prev;	//pointer to the previous device
};

typedef struct device_button_st device_button_t;

//global button list
device_button_t *button_list_head = NULL, *button_list_tail = NULL;


void add_list_member (device_button_t*);
void destroy_list_member (device_button_t*);


/*
 * init_button_list
 * Called on start or when device added or
 * removed, scans all open devices and
 * creates the global button list with
 * their button options
 *
 * Arguments: (none)
 * Returns: (none)
 */

void init_button_list () {
	const SANE_Device **devicelist;
	int currdevice, curroption, currbutton, device_has_buttons;
	SANE_Status status;
	const SANE_Option_Descriptor *tempoptdesc;
	SANE_Int numoptions;
	device_button_t *curr_device_button;

	//first, destroy the button list if it exists
	destroy_button_list();

	//get the device list
	status = sane_get_devices(&devicelist, FALSE);
	if (status != SANE_STATUS_GOOD) {
		//TODO: change the return status for this?
		fprintf(stderr, "Couldn't get device list, buttons won't be supported\n");
		return;
	}

	//now go through the list
	for (currdevice = 0; devicelist[currdevice] != NULL; currdevice++) {
		//and for each one, first create the device+button struct
		curr_device_button = (device_button_t *)malloc(sizeof(device_button_t));
		//NOTE: Can we depend on the contents of device_button_t->buttonlist being 0?
		for (currbutton = 0; currbutton < MAX_BUTTONS; currbutton++)
			curr_device_button->buttonlist[currbutton] = 0;
		add_list_member(curr_device_button);
		//then open the device
		status = sane_open(devicelist[currdevice]->name, &(curr_device_button->handle));
		if (status != SANE_STATUS_GOOD) {
			//if it didn't work, destroy its entry
			sane_close(curr_device_button->handle);
			destroy_list_member(curr_device_button);
			fprintf(stderr, "Couldn't open device %s\n", devicelist[currdevice]->name);
			continue;
		}
		device_has_buttons = 0;

		//then check all the options
		status = sane_control_option(curr_device_button->handle, 0,
			SANE_ACTION_GET_VALUE, &numoptions, NULL);
		if (status != SANE_STATUS_GOOD) {
			//if we couldn't get the number of options, skip this device
			sane_close(curr_device_button->handle);
			destroy_list_member(curr_device_button);
			fprintf(stderr, "Couldn't get number of options for device %s\n",
				devicelist[currdevice]->name);
			continue;
		}

		for (curroption = 0; curroption < numoptions; curroption++) {
			//for each option, get its name
			tempoptdesc = sane_get_option_descriptor(curr_device_button->handle, curroption);
			//if it's a button,
			//(ignore buttons that aren't booleans)
			if (!strncmp(tempoptdesc->name, "button", 6) &&
				tempoptdesc->type == SANE_TYPE_BOOL) {

				#ifdef DEBUG
					printf("Found button %s for device %s\n",
						tempoptdesc->name, devicelist[currdevice]->name);
				#endif

				device_has_buttons = 1;
				//find an empty space in the device's button list
				for (currbutton = 0; currbutton < MAX_BUTTONS; currbutton++) {
					if (curr_device_button->buttonlist[currbutton] == 0)
						break;
				}
				if (currbutton == MAX_BUTTONS) {
					fprintf(stderr, "Too many buttons for device %s\n", devicelist[currdevice]->name);
					continue;
				}

				//and add the option number there
				curr_device_button->buttonlist[currbutton] = curroption;
			}
		}

		//if the device doesn't have any options, remove it
		//NOTE: this is a bit inefficient - it would be better to scan all the options FIRST,
		//		then add the device to the list only after we know it has some buttons
		if (!device_has_buttons) {

			#ifdef DEBUG
				printf("Device %s doesn't have any buttons\n",
					devicelist[currdevice]->name);
			#endif

			sane_close(curr_device_button->handle);
			destroy_list_member(curr_device_button);
		}
	}
}


/*
 * destroy_button_list
 * Closes all open devices and frees the
 * global button list
 *
 * Arguments: (none)
 * Returns: (none)
 */

void destroy_button_list () {
	device_button_t *curr_device_button, *next_device_button;

	if (button_list_head) {
		curr_device_button = button_list_head;
		while (curr_device_button != NULL) {
			next_device_button = curr_device_button->next;
			sane_close(curr_device_button->handle);
			free(curr_device_button);
			curr_device_button = next_device_button;
		}
		button_list_head = button_list_tail = NULL;
	}
}


/*
 * scan_buttons
 * Called periodically to scan buttons
 * of existing devices to see if they've
 * been pressed, sends appropriate
 * D-BUS signals if they have
 *
 * Arguments: (none)
 * Returns: (none)
 */

void scan_buttons (DBusConnection *conn) {
	SANE_Bool buttonval;
	device_button_t *curr_device_button, *next_device_button;
	int currbutton;
	SANE_Status status;
	DBusMessage *msg;
	const SANE_Option_Descriptor *tempoptdesc;

	curr_device_button = button_list_head;

	//go through the button list until the end
	while (curr_device_button) {
		next_device_button = curr_device_button->next;
		//for each device,
		//go through the list of buttons
		for (currbutton = 0; currbutton < MAX_BUTTONS; currbutton++) {
			//stop when we've hit the end of the list
			if (curr_device_button->buttonlist[currbutton] == 0)
				break;

			//for each button, try to get the value
			status = sane_control_option(curr_device_button->handle,
				curr_device_button->buttonlist[currbutton], SANE_ACTION_GET_VALUE,
				(void *)&buttonval, NULL);
			//if the read didn't succeed, assume that there's an I/O
			// problem and remove the device from the list
			if (status != SANE_STATUS_GOOD) {
				fprintf(stderr, "Error reading value from button, removing device\n");
				sane_close(curr_device_button->handle);
				destroy_list_member(curr_device_button);
			}
			//otherwise, if the read succeeded and the button's pressed,
			if (buttonval == SANE_TRUE) {
				#ifdef DEBUG
					printf("Button pressed\n");
				#endif

				//get the button's name
				//NOTE: (shouldn't need to check the return value here to see if
				//		if succeeded because it must have read fine a few lines back)
				tempoptdesc = sane_get_option_descriptor(curr_device_button->handle,
					curr_device_button->buttonlist[currbutton]);
				//and send a signal to notify anyone who's listening
				msg = dbus_message_new_signal("/org/freedesktop/ImageCapture",
					"org.freedesktop.ImageCaptureInterface", "ButtonPressed");
				dbus_message_append_args(msg, DBUS_TYPE_STRING,
					&(tempoptdesc->name), DBUS_TYPE_INVALID);
				if (!dbus_connection_send(conn, msg, NULL))
					fprintf(stderr, "Couldn't send signal for some reason...\n");
				dbus_message_unref(msg);
			}
		}
		curr_device_button = next_device_button;
	}

	dbus_connection_flush(conn);
}


/*
 * add_list_member
 * Adds a device+button struct to end of the
 * global button list
 *
 * Arguments: device button struct (device_button_t*)
 * Returns: (none)
 */

void add_list_member (device_button_t *device_button) {

	device_button->next = NULL;
	if (button_list_tail)
		button_list_tail->next = device_button;
	device_button->prev = button_list_tail;
	button_list_tail = device_button;
	if (!button_list_head)
		button_list_head = device_button;
}


/*
 * destroy_list_member
 * Removes a device+button struct from the
 * global button list
 *
 * Arguments: device button struct (device_button_t*)
 * Returns: (none)
 */

void destroy_list_member (device_button_t *device_button) {

	if (device_button->prev)
		device_button->prev->next = device_button->next;
	if (device_button->next)
		device_button->next->prev = device_button->prev;

	if (device_button == button_list_head)
		button_list_head = device_button->next;
	if (device_button == button_list_tail)
		button_list_tail = device_button->prev;

	free(device_button);
}
