/*
	Image Capture Daemon - a D-BUS service for interfacing to SANE
	Copyright (C) 2007  Donald Straney <burntfuse@gmail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sane/sane.h>
#include <dbus/dbus.h>

#include "imcapd.h"
#include "jobs.h"

void send_reply_failure (DBusMessage *reply, dbus_uint32_t errcode) {

	//add the error code to the reply message
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &errcode, DBUS_TYPE_INVALID);
}


/*
 * GetDeviceList
 * Returns names of detected SANE devices
 *
 * D-BUS Arguments: (none)
 * D-BUS Reply: status (uint32), device 1 (string), device 2 (string)...
 */

void method_get_device_list (DBusMessage *msg, DBusMessage *reply) {
	DBusMessageIter args;
	const SANE_Device **devicelist;
	dbus_uint32_t status;
	int currdevice;

	//doesn't take any args, so we're not even going to check
	//get the device list from SANE
	status = sane_get_devices(&devicelist, FALSE);
	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(reply, status);
		return;
	}

	//init the args for the reply
	dbus_message_iter_init_append(reply, &args);
	//first, the required status
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &status);
	/* now go through the list of devices, and add each one's name to the reply.
		an int ID for each would be better, but for now it's fine since the name
		is supposed to be unique */
	for (currdevice = 0; devicelist[currdevice] != NULL; currdevice++) {
		dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING,
			&(devicelist[currdevice]->name));
	}
}


/*
 * OpenDevice
 * Opens connection to device, returns handle for job
 *
 * D-BUS Arguments: device name (string)
 * D-BUS Reply: status (uint32), job handle (uint32)
 */

void method_open_device (DBusMessage *msg, DBusMessage *reply) {
	char *devicename;
	int extjobhandle;
	dbus_uint32_t status_dbus, status;

	//check args
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING,
		&devicename, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL); 
		return;
	}

	//now, try to open a new connection to the device
	status = create_job(devicename, &extjobhandle);
	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(reply, status);
		return;
	}

	//if it worked, reply with a "success" status code and the handle
	status_dbus = SANE_STATUS_GOOD;
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
		DBUS_TYPE_UINT32, &extjobhandle, DBUS_TYPE_INVALID);
}


/*
 * StartScan
 * Starts scan for a scan job
 *
 * D-BUS Arguments: job handle (uint32)
 * D-BUS Reply: status (uint32)
 */

void method_start_scan (DBusMessage *msg, DBusMessage *reply) {
	dbus_uint32_t extjobhandle, status;

	//send a 'failed' return code if it didn't have any args
	// or if they're the wrong type
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL); 
		return;
	}

	//now, try to start the scan
	status = start_job(extjobhandle);
	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(reply, status);
		return;
	}

	//if it worked, reply with a "success" status code and the handle
	dbus_message_append_args(reply, DBUS_TYPE_UINT32,
		&status, DBUS_TYPE_INVALID);
}


/*
 * StopScan
 * Stops a job, job handle is
 * invalid after this is called
 *
 * D-BUS Arguments: job handle (uint32)
 * D-BUS Reply: status (uint32)
 */

void method_stop_scan (DBusMessage *msg, DBusMessage *reply) {
	dbus_uint32_t extjobhandle, status;

	//make sure it's passing a job handle
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL); 
		return;
	}

	//then try to stop the scan
	status = stop_job(extjobhandle);

	//and send back the status code
	dbus_message_append_args(reply, DBUS_TYPE_UINT32,
		&status, DBUS_TYPE_INVALID);
}


/*
 * ReadImage
 * Temporary hack to read from device and dump to file
 *
 * D-BUS Arguments: job handle (uint32)
 * D-BUS Reply: status (uint32)
 */

void method_read_image (DBusMessage *msg, DBusMessage *reply) {
	dbus_uint32_t extjobhandle, status;

	//make sure it's passing a job handle
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL); 
		return;
	}

	//then read the image
	status = job_read_image(extjobhandle);

	//and send back the status code
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status, DBUS_TYPE_INVALID);
}


/*
 * GetOptionInfo
 * Returns type, range, and constraints on
 * option for scan job
 *
 * D-BUS Arguments: job handle (uint32), option number (uint32)
 * D-BUS Reply: status (uint32), option name (string), option title (string),
 *	option description (string), option type (uint32), option units (uint32),
 *	option capabilities (uint32), option constraint type (uint32),
 *	[allowed value 1, allowed value 2... | min (int32), max (int32), step (int32)]
 */

void method_get_option_info (DBusMessage *msg, DBusMessage *reply) {
	dbus_uint32_t extjobhandle, option, status, temp_dbusuint;
	dbus_int32_t temp_dbusint;
	int i, numints;
	const SANE_Option_Descriptor *optiondesc;
	SANE_Handle intjobhandle;

	//get the arguments
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_UINT32, &option, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//(match option numbering)
	option++;

	intjobhandle = get_int_job_handle(extjobhandle);
	if (!intjobhandle) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//try to get the option descriptor
	optiondesc = sane_get_option_descriptor(intjobhandle, option);
	if (!optiondesc) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//now, add all the required args to the reply
	status = SANE_STATUS_GOOD;
	//yeah, it's a bit messy, but it's better than having to assume all the SANE and
	// D-BUS types are the same size
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
		DBUS_TYPE_STRING, &(optiondesc->name), DBUS_TYPE_STRING,
		&(optiondesc->title), DBUS_TYPE_STRING, &(optiondesc->desc), DBUS_TYPE_INVALID);
	temp_dbusuint = optiondesc->type;
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &temp_dbusuint, DBUS_TYPE_INVALID);
	temp_dbusuint = optiondesc->unit;
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &temp_dbusuint, DBUS_TYPE_INVALID);
	temp_dbusuint = optiondesc->cap;
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &temp_dbusuint, DBUS_TYPE_INVALID);
	temp_dbusuint = optiondesc->constraint_type;
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &temp_dbusuint, DBUS_TYPE_INVALID);

	//see if we've got to add a list of allowed values too
	switch (optiondesc->constraint_type) {
		case SANE_CONSTRAINT_RANGE:
			//if the option's got a certain range, add the range data to the reply
			temp_dbusint = optiondesc->constraint.range->min;
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &temp_dbusint, DBUS_TYPE_INVALID);
			temp_dbusint = optiondesc->constraint.range->max;
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &temp_dbusint, DBUS_TYPE_INVALID);
			temp_dbusint = optiondesc->constraint.range->quant;
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &temp_dbusint, DBUS_TYPE_INVALID);
			break;
		case SANE_CONSTRAINT_WORD_LIST:
			//if there's a list of allowed ints, first get the number of them
			numints = optiondesc->constraint.word_list[0];
			//then add them all to the reply
			for (i = 0; i < numints; i++) {
				temp_dbusint = optiondesc->constraint.word_list[i];
				dbus_message_append_args(reply, DBUS_TYPE_INT32,
					&temp_dbusint, DBUS_TYPE_INVALID);
			}
			break;
		case SANE_CONSTRAINT_STRING_LIST:
			//if there's a list of allowed strings, loop through them
			// and add until the NULL terminator
			for (i = 0; optiondesc->constraint.string_list[i] != NULL; i++) {
				dbus_message_append_args(reply, DBUS_TYPE_STRING,
					&(optiondesc->constraint.string_list[i]), DBUS_TYPE_INVALID);
			}
			break;

		//with no constraints, we don't have to do anything
		default:
			break;	//put here to make gcc happy
	}
}


/*
 * GetNumOptions
 * Returns the number of options
 * for a certain job
 *
 * D-BUS Arguments: job handle (uint32)
 * D-BUS Return: status (uint32), number of options (uint32)
 */

void method_get_num_options (DBusMessage *msg, DBusMessage *reply) {
	SANE_Int numoptions_sane;
	dbus_uint32_t extjobhandle, status, numoptions;
	SANE_Handle intjobhandle;

	//get the job handle
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	intjobhandle = get_int_job_handle(extjobhandle);
	if (!intjobhandle) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//get the first option, which contains the number of options
	status = sane_control_option(intjobhandle, 0, SANE_ACTION_GET_VALUE,
		(void *)&numoptions_sane, NULL);
	/* the number of options includes the "size" one which we just used,
		but we want to return the number of valid options (this one's
		hidden from the outside world), so return one less than that */
	numoptions = numoptions_sane - 1;

	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
		DBUS_TYPE_UINT32, &numoptions, DBUS_TYPE_INVALID);
}


/*
 * SetOptionValue
 * Sets the value of an option
 * for a certain job
 *
 * D-BUS Arguments: job handle (uint32), option number (uint32), option value (int32 | string | bool)
 * D-BUS Return: status (uint32)
 */

void method_set_option_value (DBusMessage *msg, DBusMessage *reply) {
	DBusMessageIter args;
	dbus_uint32_t extjobhandle, optionnum, status;
	dbus_int32_t value_int_dbus;
	SANE_Int value_int_sane;
	char *value_string;
	dbus_bool_t value_bool_dbus;
	SANE_Bool value_bool_sane;
	int optiontype;
	SANE_Handle intjobhandle;
	void *value;

	//get the arguments
	dbus_message_iter_init(msg, &args);
	//first the fixed ones,
	if (dbus_message_iter_get_arg_type(&args) == DBUS_TYPE_UINT32) {
		dbus_message_iter_get_basic(&args, (void *)&extjobhandle);
	} else {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	intjobhandle = get_int_job_handle(extjobhandle);
	if (!intjobhandle) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	if (dbus_message_iter_next(&args) &&
		dbus_message_iter_get_arg_type(&args) == DBUS_TYPE_UINT32) {

		dbus_message_iter_get_basic(&args, (void *)&optionnum);
	} else {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//(match the option number with SANE's internal option numbering)
	optionnum++;

	//then the variable-type option value
	if (!dbus_message_iter_next(&args)) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	optiontype = dbus_message_iter_get_arg_type(&args);
	if (optiontype == DBUS_TYPE_INT32) {
		dbus_message_iter_get_basic(&args, (void *)&value_int_dbus);
		value_int_sane = value_int_dbus;
		value = (void *)&value_int_sane;
	} else if (optiontype == DBUS_TYPE_STRING) {
		value = (void *)&value_string;
		dbus_message_iter_get_basic(&args, value);
	} else if (optiontype == DBUS_TYPE_BOOLEAN) {
		dbus_message_iter_get_basic(&args, (void *)&value_bool_dbus);
		value_bool_sane = value_bool_dbus;
		value = (void *)&value_bool_sane;
	//TODO: figure out how to handle other option types!
	} else {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	status = sane_control_option(intjobhandle, optionnum, SANE_ACTION_SET_VALUE,
		value, NULL);

	dbus_message_append_args(reply, DBUS_TYPE_UINT32,
		&status, DBUS_TYPE_INVALID);
}


/*
 * GetOptionValue
 * Gets the value of an option
 * for a certain job
 *
 * D-BUS Arguments: job handle (uint32), option number (uint32)
 * D-BUS Return: status (uint32), option value (int32 | string | bool)
 */

void method_get_option_value (DBusMessage *msg, DBusMessage *reply) {
	dbus_uint32_t extjobhandle, optionnum, status;
	dbus_int32_t retval_int;
	dbus_bool_t retval_bool;
	SANE_Handle intjobhandle;
	const SANE_Option_Descriptor *optiondesc;

	//get the args
	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT32,
		&extjobhandle, DBUS_TYPE_UINT32, &optionnum, DBUS_TYPE_INVALID)) {

		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	intjobhandle = get_int_job_handle(extjobhandle);
	if (!intjobhandle) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//match option numbering
	optionnum++;

	//get the option descriptor for later use
	optiondesc = sane_get_option_descriptor(intjobhandle, optionnum);
	if (!optiondesc) {
		send_reply_failure(reply, SANE_STATUS_INVAL);
		return;
	}

	//and try to get the option's value
	char dataptr[optiondesc->size];
	status = sane_control_option(intjobhandle, optionnum,
		SANE_ACTION_GET_VALUE, (void *)dataptr, NULL);

	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(reply, status);
		return;
	}

	//I'm hoping this converts between the SANE and D-BUS types correctly
	// and won't break if they have different sizes...
	switch (optiondesc->type) {
		case SANE_TYPE_INT:
			retval_int = *((SANE_Int *)dataptr);
			dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
				DBUS_TYPE_INT32, &retval_int, DBUS_TYPE_INVALID);
			break;
		case SANE_TYPE_STRING:
			dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
				DBUS_TYPE_STRING, &dataptr, DBUS_TYPE_INVALID);
			break;
		case SANE_TYPE_BOOL:
			retval_bool = *((SANE_Bool *)dataptr);
			dbus_message_append_args(reply, DBUS_TYPE_UINT32, &status,
				DBUS_TYPE_BOOLEAN, &retval_bool, DBUS_TYPE_INVALID);
			break;
	//TODO: handle other options somehow!!!
		default:
			//right now, other option types just return "unsupported",
			// because I don't know what they do
			send_reply_failure(reply, SANE_STATUS_UNSUPPORTED);
			return;
	}
}
