/*
	Image Capture Daemon - a D-BUS service for interfacing to SANE
	Copyright (C) 2007  Donald Straney <burntfuse@gmail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <sane/sane.h>

#include "jobs.h"
#include "imcapd.h"

//stores handles for active jobs
SANE_Handle jobhandles[NUM_SIMULT_SCANS];
//timeout counters for active jobs
int jobtimeouts[NUM_SIMULT_SCANS];


/*
 * create_job
 * Opens a connection to the specified device
 *
 * Arguments: device name (string), pointer to job handle (int*)
 * Returns: SANE_STATUS_GOOD if successful
 */

SANE_Status create_job (const char *devicename, int *job) {
	SANE_Handle devicehandle;
	SANE_Status status;
	int extjobhandle;

	#ifdef DEBUG
		printf("Trying to open device %s\n", devicename);
	#endif

	//try to open the device
	status = sane_open(devicename, &devicehandle);
	if (status != SANE_STATUS_GOOD)
		return status;

	//store the job handle in our SANE_Handle table
	//and pass back an int to the caller to keep track of it
	extjobhandle = 0;
	while (jobhandles[extjobhandle] != NULL) {
		extjobhandle++;
		//prevent overflow if the table's already full
		if (extjobhandle == NUM_SIMULT_SCANS) {
			fprintf(stderr, "Too many scan jobs open at once\n");
			return SANE_STATUS_DEVICE_BUSY;
		}
	}

	#ifdef DEBUG
		printf("Opened device %s, returning new job handle %d\n",
		devicename, extjobhandle);
	#endif

	//once we've found an empty slot, store the handle there
	jobhandles[extjobhandle] = devicehandle;
	//reset its timeout counter
	jobtimeouts[extjobhandle] = 0;
	//and return it to the caller
	if (job)
		*job = extjobhandle;

	return SANE_STATUS_GOOD;
}


/*
 * start_job
 * Starts a scan for the specified job
 *
 * Arguments: job handle (int)
 * Returns: SANE_STATUS_GOOD if successful
 */

SANE_Status start_job (int job) {
	SANE_Status status;
	SANE_Handle intjobhandle;

	//get the SANE device handle
	intjobhandle = get_int_job_handle(job);
	if (!intjobhandle)
		return SANE_STATUS_INVAL;

	#ifdef DEBUG
		printf("Starting scan for job %d...\n", job);
	#endif

	//then start the scan
	status = sane_start(intjobhandle);

	return status;
}


/*
 * stop_job
 * Stops the specified job
 *
 * Arguments: job handle (int)
 * Returns: SANE_STATUS_GOOD if successful
 */

SANE_Status stop_job (int job) {
	SANE_Handle intjobhandle;

	//make sure the job handle's valid...
	intjobhandle = get_int_job_handle(job);
	if (!intjobhandle)
		return SANE_STATUS_INVAL;

	#ifdef DEBUG
		printf("Stopping scan for job %d...\n", job);
	#endif

	//and cancel the scan
	sane_close(intjobhandle);	//what's the difference between this and sane_cancel?
	jobhandles[job] = NULL;

	return SANE_STATUS_GOOD;
}


/*
 * job_read_image
 * Reads an image for the specified job,
 * places in /tmp/scanned_image
 *
 * Arguments: job handle (int)
 * Returns: SANE_STATUS_GOOD if successful
 */

SANE_Status job_read_image (int job) {
	SANE_Handle intjobhandle;
	FILE *tmpfile;
	SANE_Status status;
	int datalen;
	char databuf[BUF_SIZE];

	/* NOTE: this is just a temporary routine to test the image capture:
		the final version will have a much faster/more flexible way
		for passing back the image.
		for now, be careful since the file gets overwritten with
		each call to ReadImage, so two clients accessing this at the
		same time will probably start stepping on each other's toes. */

	//retrieve our SANE scan handle
	intjobhandle = get_int_job_handle(job);
	if (!intjobhandle)
		return SANE_STATUS_INVAL;

	//also, reset the job timeout to show it's being used
	jobtimeouts[job] = 0;

	/* now, we're going to read in the image and dump
		the raw data to a temporary file, then pass back
		its name to the calling program */

	//open the file
	tmpfile = fopen(OUT_FILENAME, "w");
	if (!tmpfile)
		return SANE_STATUS_NO_MEM;

	//now read from the device until the end of the image
	status = SANE_STATUS_GOOD;
	while (status == SANE_STATUS_GOOD) {
		status = sane_read(intjobhandle, databuf, BUF_SIZE, &datalen);
		fwrite(databuf, 1, datalen, tmpfile);
	}
	//TODO: have it time out if it doesn't read any data for a certain amount of time

	//finally, close the file
	fclose(tmpfile);

	//if the return code was EOF, transfer was successful,
	// so set status to show that
	if (status == SANE_STATUS_EOF)
		status = SANE_STATUS_GOOD;

	return status;
}


/*
 * check_job_timeouts
 * Runs job timeout counters, checks if any jobs
 * have timed out, and closes any that have
 *
 * Arguments: (none)
 * Returns: (none)
 */

void check_job_timeouts () {
	int currjob;

	//go through the job list
	for (currjob = 0; currjob < NUM_SIMULT_SCANS; currjob++) {
		//only check active jobs
		if (jobhandles[currjob]) {
			//increment the timeout counter for this job
			jobtimeouts[currjob] += SIG_CHECK_INT;
			//if it's past the timeout value, then close the job
			if (jobtimeouts[currjob] >= JOB_TIMEOUT) {
				#ifdef DEBUG
					printf("Job %d timed out\n", currjob);
				#endif
				stop_job(currjob);
			}
		}
	}
}


/*
 * get_int_job_handle
 * Gets the internal job handle from
 * the external job handle, with
 * proper checking
 *
 * Arguments: job handle (int)
 * Returns: internal job handle, or NULL on error
 */

SANE_Handle get_int_job_handle (int job) {

	//check the range
	if (job >= NUM_SIMULT_SCANS || job < 0)
		return NULL;

	return jobhandles[job];
}
