/*
	Image Capture Daemon - a D-BUS service for interfacing to SANE
	Copyright (C) 2007  Donald Straney <burntfuse@gmail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <dbus/dbus.h>
#include <sane/sane.h>

#include "imcapd.h"
#include "dbus-methods.h"
#include "jobs.h"
#include "button-notify.h"

//table of dbus methods with functions to handle them
void *methodtable[] =
	{"GetDeviceList", method_get_device_list, "OpenDevice", method_open_device,
	"StartScan", method_start_scan,	"StopScan", method_stop_scan,
	"ReadImage", method_read_image, "GetOptionInfo", method_get_option_info,
	"GetNumOptions", method_get_num_options, "SetOptionValue", method_set_option_value,
	"GetOptionValue", method_get_option_value, NULL};
//used to stop main loop on signal
int run = 1;
//time before rescanning devices to build button list - set to -1 if no rescan pending
int rescan_counter = -1;


void handle_signal (int signal) {
	//if we caught a SIGINT or SIGTERM, stop the main loop
	#ifdef DEBUG
		printf("Caught signal, exiting...\n");
	#endif
	run = 0;
}


DBusHandlerResult handle_request (DBusConnection *conn, DBusMessage *msg, void *userdata) {
	DBusMessage *reply;

	//look through the method table to see if we can handle this message
	int i = 0;
	while (TRUE) {
		//make sure we're not at the end of the table
		if (methodtable[i] == NULL) {
			#ifdef DEBUG
				printf("Got unrecognized method call\n");
			#endif
			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
		//if not, check if this call matches the received one
		} else if (dbus_message_is_method_call(msg,
			"org.freedesktop.ImageCaptureInterface", methodtable[i])) {

			//if it does, call the appropriate method
			#ifdef DEBUG
				printf("Method %s called\n", (char *)methodtable[i]);
			#endif
			//but first, create a reply message for it to use
			reply = dbus_message_new_method_return(msg);
			void (*callback)(DBusMessage*, DBusMessage*) =
				(void (*)(DBusMessage*, DBusMessage*))methodtable[++i];
			//TODO: create new thread for function so long I/O operations don't block requests
			// will have to be careful about jobs timing out during read (reset in read loop?)
			callback(msg, reply);
			//once the method's done, send the reply
			dbus_connection_send(conn, reply, NULL);
			dbus_connection_flush(conn);
			dbus_message_unref(reply);
			break;
		}
		//if it doesn't match, look at the next one
		i += 2;
	}
	return DBUS_HANDLER_RESULT_HANDLED;
}


DBusHandlerResult device_removed_added (DBusConnection *conn, DBusMessage *msg, void *data) {

	//if we got a signal from HAL, see if a device has been added or removed
	if (dbus_message_is_signal(msg, "org.freedesktop.Hal.Manager", "DeviceAdded") ||
		dbus_message_is_signal(msg, "org.freedesktop.Hal.Manager", "DeviceRemoved")) {

		#ifdef DEBUG
			printf("Device added or removed\n");
		#endif
		/* if so, then wait for a bit before re-scanning.
			the logic for this is that since one piece of
			hardware corresponds to more than one object
			in HAL most of the time, it's a waste to re-scan
			4 times just because a USB mouse was unplugged.
			the scanner doesn't need to be instantly available
			for button presses anyways, so we can just wait
			for half a second and it'll save a lot of trouble. */

		//so init the timeout counter - the main loop will take care of the rest
		rescan_counter = RESCAN_TIMEOUT;
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}


int main (int argc, char **argv) {

	DBusConnection *conn;
	DBusError error;

	//init SANE
	if (sane_init(NULL, NULL) != SANE_STATUS_GOOD) {
		fprintf(stderr, "Couldn't init SANE!\n");
		exit(1);
	}

	dbus_error_init(&error);
	//connect to DBUS
	conn = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
	if (!conn) {
		fprintf(stderr, "Couldn't connect to DBUS!\n");
		if (dbus_error_is_set(&error)) {
			fprintf(stderr, "%s\n", error.message);
			dbus_error_free(&error);
		}
		sane_exit();
		exit(1);
	}

	//try to get a name
	if (dbus_bus_request_name(conn, "org.freedesktop.ImageCapture",
		DBUS_NAME_FLAG_REPLACE_EXISTING, &error) !=
		DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {

		fprintf(stderr, "Couldn't lock name!\n");
		if (dbus_error_is_set(&error)) {
			fprintf(stderr, "%s\n", error.message);
			dbus_error_free(&error);
		}
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//export an object for requests
	DBusObjectPathVTable vtable;
	vtable.message_function = handle_request;
	vtable.unregister_function = NULL;
	if (!dbus_connection_register_object_path(conn,
		"/org/freedesktop/ImageCapture", &vtable, NULL)) {

		fprintf(stderr, "Couldn't export object!\n");
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//listen for HAL messages,
	dbus_bus_add_match(conn,
		"type='signal',interface='org.freedesktop.Hal.Manager'", &error);
	if (dbus_error_is_set(&error)) {
		fprintf(stderr, "Couldn't add match!\n%s\n", error.message);
		dbus_error_free(&error);
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}
	if (!dbus_connection_add_filter(conn, device_removed_added, NULL, NULL)) {
		fprintf(stderr, "Couldn't add filter!\n");
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}
	dbus_connection_flush(conn);

	//install a signal handler so the daemon can exit cleanly,
	struct sigaction sighandler;
	sighandler.sa_handler = handle_signal;
	if (sigaction(SIGINT, &sighandler, NULL) == -1 ||
		sigaction(SIGTERM, &sighandler, NULL) == -1) {

		fprintf(stderr, "Couldn't install signal handler, continuing...\n");
	}

	//get the initial list of scanner buttons
	init_button_list();

	//and do nothing until we get an exit signal
	while (dbus_connection_read_write_dispatch(conn, SIG_CHECK_INT) && run) {
		//...except for waking up to check if any job handles have timed out
		check_job_timeouts();
		//...and to scan the buttons
		scan_buttons(conn);

		//if we were supposed to rebuild the button list, do that too
		if (rescan_counter > 0 && (rescan_counter -= SIG_CHECK_INT) <= 0) {
			#ifdef DEBUG
				printf("Rebuilding button list\n");
			#endif
			init_button_list();
		}
	}

	destroy_button_list();
	dbus_connection_unref(conn);
	sane_exit();
	return 0;
}
